`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    14:13:52 03/09/2007 
// Design Name: 
// Module Name:    serialmult 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: Two complement 16x16 multiplier
// If should be noted that I am only using the upper 16 bits of the product.
// To get all 32 bits, simply change the dimensions on P to [31:0]
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module serialmult(C, R,A,B, P, L, D);
    input C;				//system clock
    input R;				//system reset
    input [15:0] A,B;		//multiplier
	 output [15:0] P;		//product
    input L;				//load multiplier
	output D;				//done flag
	
	wire [14:0] cn;			//carry bits
	wire E,Add,S;
	reg [15:0] Breg;
	reg En;
	reg [5:0] cntr;
	reg [15:0] P;
	reg D;

//--------------------------------------------
// Done flag
// D =1 when product is complete
// D = 0 when multiplier is busy
//
// This flag is cleared to 0 when you load
// the multiplier registers (L)
//--------------------------------------------
always @ (posedge C or posedge R)
	if(R)
		D <= 0;
	else if (L)	//indicate busy when mult is loaded
		D <= 0;
	else if (cntr == 0)	//inidcate done when all bits shifted out
		D <= 1;

//-------------------------------------------
// Enable multiplication
// This bit also indicates busy
//-------------------------------------------

always @ (posedge C or posedge R)
	if(R)
		En <= 0;
	else if (L)
		En <= 1;	//do multiply
	else if (cntr == 0)
		En <= 0;	//multiply done

//--------------------------------------------
//		Shift counter
//	Keeps track of the number of shifts
// It takes 33 clocks to do a 16x16 mulitiply
//--------------------------------------------
always @ (posedge C or posedge R)
	if(R)
		cntr <= 0;
	else if (L)
		cntr <= 32;	//load count-1
	else if (En)	//while multipling
		cntr <= cntr - 1;	//decrement shift count

//--------------------------------------------
// product shift register
// This is only the upper 16 bits of the product
//--------------------------------------------
always @ (posedge C or posedge R)
		if(R)
			P <= 0;
		else if (En)	//if multiply enabled
			P <= {S,P[15:1]};	//shift in bit from
									//multiplier hardware

//---------------------------------------------
// B register
// This register needs to do a sign extend
// as B is shifted into the multiplier
//---------------------------------------------
always @ (posedge C or posedge R)
	if(R)
		Breg <= 0;
	else if (L)
		Breg <= B;	//load multiplier
	else if (En)	//while multipling...
		Breg <= {Breg[15],Breg[15:1]};	//arithmetic shift right

//--------------------------------------------
// add/subtract control
//--------------------------------------------		
addsubctl addsubctl0 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .I(Breg[0]),
	 .en(En),
    .E(E), 
    .A(Add)
    );

//--------------------------------------------
// multiplier cells
//--------------------------------------------
multcell Mu15 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[15]), 
    .L(L), 
    .I(cn[0]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[0])
    );

multcell Mu14 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[14]), 
    .L(L), 
    .I(cn[0]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[1])
    );
	 
multcell Mu13 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[13]), 
    .L(L), 
    .I(cn[1]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[2])
    );
	 
multcell Mu12 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[12]), 
    .L(L), 
    .I(cn[2]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[3])
    );
	 
multcell Mu11 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[11]), 
    .L(L), 
    .I(cn[3]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[4])
    );
	 
multcell Mu10 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[10]), 
    .L(L), 
    .I(cn[4]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[5])
    );
	 
multcell Mu9 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[9]), 
    .L(L), 
    .I(cn[5]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[6])
    );

multcell Mu8 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[8]), 
    .L(L), 
    .I(cn[6]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[7])
    );

multcell Mu7 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[7]), 
    .L(L), 
    .I(cn[7]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[8])
    );
	 
multcell Mu6 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[6]), 
    .L(L), 
    .I(cn[8]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[9])
    );
	 
multcell Mu5 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[5]), 
    .L(L), 
    .I(cn[9]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[10])
    );
	 
multcell Mu4 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[4]), 
    .L(L), 
    .I(cn[10]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[11])
    );
	 
multcell Mu3 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[3]), 
    .L(L), 
    .I(cn[11]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[12])
    );
	 
multcell Mu2 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[2]), 
    .L(L), 
    .I(cn[12]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[13])
    );

multcell Mu1 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[1]), 
    .L(L), 
    .I(cn[13]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(cn[14])
    );

multcell Mu0 (
    .C(C), 
    .R(R), 
	 .sC(L),
    .M(A[0]), 
    .L(L), 
    .I(cn[14]), 
    .A(Add), 
    .E(E), 
	 .en(En),
    .O(S)
    );

endmodule
